#include "cluster.hh"
#include "functions.hh"

using namespace std;

Cluster::~Cluster() {}

void Cluster::free_rec() {
    for (cluster_ptrs_t::iterator it=children_.begin(); children_.end()!=it; ++it) {
	(*it)->free_rec();
	delete(*it);
    }
    for (func_ptrs_t::iterator it=funcs_.begin(); funcs_.end()!=it; ++it) {
	delete(*it);
    }
}


void Cluster::print_debug() {
    cout <<"CLUSTER: "<<name_<<endl;
    cout <<"--------------------"<<endl;
    for (var_ptrs_t::iterator it=vars_.begin(); vars_.end()!=it; ++it) {
	(*it)->print_debug();
    }
    for (func_ptrs_t::iterator it=funcs_.begin(); funcs_.end()!=it; ++it) {
	(*it)->print_debug();
    }
    for (func_ptrs_t::iterator it=cons_.begin(); cons_.end()!=it; ++it) {
	(*it)->print_debug();
    }
    cout << "Children:";
    for (cluster_ptrs_t::iterator it=children_.begin(); children_.end()!=it; ++it) {
	cout << " " << (*it)->name_;
    }
    cout << endl;
}

void Cluster::print_debug_rec() {
    print_debug();cout<<endl;
    for (cluster_ptrs_t::iterator it=children_.begin(); children_.end()!=it; ++it) {
	(*it)->print_debug_rec();
    }
}



void Cluster::ask_children() {
    // ask each child
    if (VERBOSE) cout <<"Cluster "<<name_<<": ask children"<<endl;

    for (cluster_ptrs_t::iterator it=children_.begin(); children_.end()!=it; ++it) {
	funcs_.push_back((*it)->marginalize());
    }
    
    if (VERBOSE) {
	cout <<endl<< "Children of "<<name_<<" answered."<<endl;
	print_debug();
    }

}

bool Cluster::propagate_to(Variable *v) {
    for (func_ptrs_t::iterator it=cons_.begin(); cons_.end() != it; ++it) {
	if (!(*it)->propagate_to(v)) return false;
    }
    return true;
}

bool Cluster::contains_var(Variable *v) {
    for (var_ptrs_t::iterator it=vars_.begin(); vars_.end() != it; ++it) {    
	if ( *it == v ) return true;
    }
    return false;
}


void Cluster::partition_shared_vars(var_ptrs_t &sep_vars,var_ptrs_t &elim_vars) {
    sep_vars.resize(0);
    elim_vars.resize(0);
    
    // all variables of *this that are shared with *parent_ go into sep_vars
    // all other variables of *this go into elim_vars
    for (var_ptrs_t::iterator it=vars_.begin(); vars_.end() != it; ++it) {
	if (parent_ && parent_->contains_var(*it))
	    sep_vars.push_back(*it);
	else
	    elim_vars.push_back(*it);
    }
}


MaterializedFunction *Cluster::marginalize() {
    ask_children();
    
    // determine separator variables and elimimination variables
    var_ptrs_t sep_vars;
    var_ptrs_t elim_vars;
    partition_shared_vars(sep_vars,elim_vars);
    
    if (VERBOSE) {
	cout <<"Separator Vars:"<<endl;
	for (var_ptrs_t::iterator it=sep_vars.begin(); sep_vars.end()!=it; ++it) {
	    (*it)->print_debug();
	}
	cout <<"Elimination Vars:"<<endl;
	for (var_ptrs_t::iterator it=elim_vars.begin(); elim_vars.end()!=it; ++it) {
	    (*it)->print_debug();
	} 
	cout <<"----------"<<endl;
    }

    MaterializedFunction *g = new MaterializedFunction("g_{"+name_+"}",sep_vars); // the message is a function of the separator variables
    marginalize_on(sep_vars, 0, elim_vars, g);

    return g;
}



/* enumerate all assignments for the separator variables recursively, for each
 * assignment compute the corresponding function value of f by a call to
 * maximize_over(...)
 */

void Cluster::marginalize_on(var_ptrs_t &sep_vars, uint sep_vars_idx,
			     var_ptrs_t &elim_vars, MaterializedFunction *g)
{
    if ( sep_vars_idx < sep_vars.size() ) {
	Variable *enum_var=sep_vars[sep_vars_idx]; // the variable that is enumerated
	
	Domain dom=enum_var->get_dom(); // store domain before propagating
	if (propagate_to(enum_var)) {
	    Domain pdom=enum_var->get_dom(); // propagated domain
	    if (VERBOSE) {
		cout <<"marginalize "<< enum_var->get_name() 
		     <<" in ";
		pdom.print_debug();
		cout <<endl;
	    }
	    for (int i=pdom.first(); i!=-1; i=pdom.next(i)) {
		enum_var->set_singleton(i); // set domain to the singleton domain {i}
		if (VERBOSE) enum_var->print_debug();
		marginalize_on(sep_vars,sep_vars_idx+1,elim_vars,g); // recurse
	    }
	}
	enum_var->set_dom(dom); // restore old domain of enumerated variable
    } else {
	g->set_value(
	    (elim_vars.size()>0)
	    ?
	    maximize_over(elim_vars, 0, INT_MIN)
	    :
	    sum_funcs()
	    );
    }
}


// compute the sum of function values
inline
int Cluster::sum_funcs() {
    int sum=0;
    
    if (VERBOSE) cout << "Sum ";
    for (func_ptrs_t::iterator it=funcs_.begin(); it!=funcs_.end(); ++it) {
	int val = (*it)->value();
	if (VERBOSE) cout <<"  "<< (*it)->get_name() << "=" << val;
	if (val == INT_MIN) sum = INT_MIN; else sum += val;
    }
    if (VERBOSE) cout << endl;
    // cout << "sum: "<<sum<<" max so far: "<<maximum<<endl;
    return sum;
}


/* maximize the sum of values of the functions of the cluster over all
 * assignments to the elimination variables
 */

// CALL ONLY if there is at least one elim_var, i.e.
//  pre: elim_vars_idx < elim_vars.size()

int Cluster::maximize_over(var_ptrs_t &elim_vars, uint elim_vars_idx, int maximum, vector<int> *pbest) {
    Variable *enum_var=elim_vars[elim_vars_idx]; // the variable that is enumerated
    Domain dom=enum_var->get_dom();
	
    if (propagate_to(enum_var)) {
	Domain pdom=enum_var->get_dom(); // propagated domain
	if (VERBOSE && ! enum_var->is_singleton()) {
	  cout <<"maximize "<<enum_var->get_name()
	       <<" in "; pdom.print_debug(); cout <<endl;
	}
	if (elim_vars_idx+1==elim_vars.size()) { 
	    for (int i=pdom.first(); i!=-1; i=pdom.next(i)) {
		enum_var->set_singleton(i); // set domain to the singleton domain {i}
		
		if (VERBOSE) cout << "At "<<i<<" ";
		int new_sum=sum_funcs();
		if ( new_sum > maximum ) {
		    maximum=new_sum;
		    if (pbest) {
			if (VERBOSE) cout << "New maximum "<<new_sum<<" at "<<i<<endl;
			for (uint i=0; i<elim_vars.size(); i++)
			    (*pbest)[i] = elim_vars[i]->get_singleton();
		    }
		}
	    }
	} else {
	    for (int i=pdom.first(); i!=-1; i=pdom.next(i)) {
		enum_var->set_singleton(i); // set domain to the singleton domain {i}
		if (VERBOSE) enum_var->print_debug();
		maximum = maximize_over(elim_vars, elim_vars_idx+1, maximum, pbest); // recurse
	    }
	}
    }
    enum_var->set_dom(dom); // restore old domain of enumerated variable
    return maximum;
}


void Cluster::backtrace() {
    // PRE: the separator variables are fixed

    // determine separator variables and elimimination variables
    var_ptrs_t sep_vars;
    var_ptrs_t elim_vars;
    partition_shared_vars(sep_vars,elim_vars);
    
    vector<int> best;
    best.resize(elim_vars.size());
    
    // determine optimal assignment for elimination variables
    
    if (elim_vars.size()>0)
	maximize_over(elim_vars, 0, INT_MIN, &best);
    
    // and fix the variables
    for (uint i=0; i<elim_vars.size(); i++) {
	elim_vars[i]->set_singleton(best[i]);
	if (VERBOSE) { cout <<"BACKTRACE: "; elim_vars[i]->print_debug(); }
    }
        
    // recurse for each child
    for (cluster_ptrs_t::iterator it=children_.begin(); children_.end() != it; ++it) {
	(*it)->backtrace();
    }
}
