#ifndef FUNCTIONS_HH
#define FUNCTIONS_HH

#include "variable.hh"

/*
  A function on variables
*/

class Function {
protected:
    const std::string name_; // for debugging/output

    var_ptrs_t vars_; // arguments of the function
public:
    Function(const std::string & name):name_(name) {}
    Function(const std::string & name, var_ptrs_t &vars):name_(name), vars_(vars) {}
    virtual ~Function();
    
    virtual int value() const=0; // evaluate the function for the current constraint store
    virtual bool propagate_to(Variable *v) const {return true;} // propagate the constraint
    
    virtual void print_debug() const { std::cout <<name_<<std::endl; } 
    const std::string &get_name() const { return name_; } 
};



/* Materialized Function */
class MaterializedFunction: public Function {
protected:
    std::vector<Domain> domains_; // for which domains values are stored (only bounds of domains are relevant)
    std::vector<int> store_;
    int offset_;

private:
    int index() const {
	if (vars_.size()==0) {return 0;}
	int idx=vars_[0]->get_singleton();
	
	for(uint i=1; i<domains_.size(); ++i) {
	    idx *= domains_[i].size();
	    idx += vars_[i]->get_singleton();
	}
	
	//assert(idx-offset_>=0); 
	//assert(idx-offset_<(int)store_.size());
	
	return idx-offset_;
    }

    int offset() const {
	if (vars_.size()==0) {return 0;}
	int off=domains_[0].get_lbound();
	
	for(uint i=1; i<domains_.size(); ++i) {
	    off *= domains_[i].size();
	    off += domains_[i].get_lbound();
	}
	
	return off;
    }

public:
    MaterializedFunction(const std::string &name, var_ptrs_t vars): Function(name,vars) {
	// set domains and reserve storage
	int size=1;
	domains_.resize(0);
	for(var_ptrs_t::iterator it=vars_.begin(); vars_.end() != it; ++it) {
	    domains_.push_back((*it)->get_dom());
	    size *= (*it)->size();
	}
	store_.resize(size);
	
	// fill with -infty
	for (std::vector<int>::iterator it=store_.begin(); it!=store_.end();  ++it) *it=INT_MIN;
	
	offset_=offset();
    }
    
    virtual ~MaterializedFunction();
    
    void set_value(int val) {
	int idx=index();
	if (VERBOSE) std::cout << "set value store["<<idx<<"]="<<val<<std::endl;
	store_[idx]=val;
    }
    virtual int value() const {
	return store_[index()];
    }

    int value(int idx) const {
	return store_[idx];
    }

};


#endif // FUNCTIONS_HH
